#include "CMeshLoaderSimple.h"

#include "glm/glm.hpp"
#include "glm/gtc/matrix_transform.hpp" //perspective, trans etc
#include "glm/gtc/type_ptr.hpp" //value_ptr

#include <stdio.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <math.h>
#include <iomanip>
#include <string>
#include <streambuf>

class SColor
{

public:

    float Red, Green, Blue;

    SColor()
        : Red(0.7f), Green(0.7f), Blue(0.7f)
    {}

};

class SVector3
{

public:

	float X, Y, Z;

	SVector3()
		: X(0), Y(0), Z(0)
	{}

	SVector3(float in)
		: X(in), Y(in), Z(in)
	{}

	SVector3(float in_x, float in_y, float in_z)
		: X(in_x), Y(in_y), Z(in_z)
	{}

	SVector3 crossProduct(SVector3 const & v) const
	{
		return SVector3(Y*v.Z - v.Y*Z, v.X*Z - X*v.Z, X*v.Y - v.X*Y);
	}

	float dotProduct(SVector3 const & v) const
	{
		return X*v.X + Y*v.Y + Z*v.Z;
	}

	float length() const
	{
		return sqrtf(X*X + Y*Y + Z*Z);
	}

	SVector3 operator + (SVector3 const & v) const
	{
		return SVector3(X+v.X, Y+v.Y, Z+v.Z);
	}

	SVector3 & operator += (SVector3 const & v)
	{
		X += v.X;
		Y += v.Y;
		Z += v.Z;

		return * this;
	}

	SVector3 operator - (SVector3 const & v) const
	{
		return SVector3(X-v.X, Y-v.Y, Z-v.Z);
	}

	SVector3 & operator -= (SVector3 const & v)
	{
		X -= v.X;
		Y -= v.Y;
		Z -= v.Z;

		return * this;
	}

	SVector3 operator * (SVector3 const & v) const
	{
		return SVector3(X*v.X, Y*v.Y, Z*v.Z);
	}

	SVector3 & operator *= (SVector3 const & v)
	{
		X *= v.X;
		Y *= v.Y;
		Z *= v.Z;

		return * this;
	}

	SVector3 operator / (SVector3 const & v) const
	{
		return SVector3(X/v.X, Y/v.Y, Z/v.Z);
	}

	SVector3 & operator /= (SVector3 const & v)
	{
		X /= v.X;
		Y /= v.Y;
		Z /= v.Z;

		return * this;
	}

	SVector3 operator * (float const s) const
	{
		return SVector3(X*s, Y*s, Z*s);
	}

	SVector3 & operator *= (float const s)
	{
		X *= s;
		Y *= s;
		Z *= s;

		return * this;
	}

	SVector3 operator / (float const s) const
	{
		return SVector3(X/s, Y/s, Z/s);
	}

	SVector3 & operator /= (float const s)
	{
		X /= s;
		Y /= s;
		Z /= s;

		return * this;
	}

};

class SVertex
{

public:

    SVector3 Position;
    SColor Color;

};

class CMesh
{

    friend class CVertexBufferObject;
    friend class CMeshLoader;

    struct STriangle
    {
        unsigned int VertexIndex1, VertexIndex2, VertexIndex3;
		SColor Color;
    };

    std::vector<SVertex> Vertices;
    std::vector<STriangle> Triangles;

    CMesh();

public:

    ~CMesh();

	void centerMeshByAverage(CMesh &Mesh);
	void centerMeshByExtents(SVector3 const & CenterLocation);

	void resizeMesh(SVector3 const &Scale);

};

bool CMeshLoader::loadVertexBufferObjectFromMesh(std::string const &fileName, int &TriangleCount, GLuint &PositionBufferHandle, GLuint &TriBufferHandle, GLuint &VertexNormalBufferHandle)
{
	 CMesh* Mesh = loadASCIIMesh(fileName);
	 if (! Mesh)
       return false;

    //CMesh::centerMeshByAverage(Mesh);
    float minx, maxx, miny, maxy, minz, maxz;
    minx = maxx = miny = maxy = minz = maxz = 0;
    for (int i = 0; i < Mesh->Vertices.size(); i++) {
       minx = std::min(minx, Mesh->Vertices.at(i).Position.X);
       maxx = std::max(maxx, Mesh->Vertices.at(i).Position.X);
       miny = std::min(miny, Mesh->Vertices.at(i).Position.Y);
       maxy = std::max(maxy, Mesh->Vertices.at(i).Position.Y);
       minz = std::min(minz, Mesh->Vertices.at(i).Position.Z);
       maxz = std::max(maxz, Mesh->Vertices.at(i).Position.Z);
    }
    SVector3* offset = new SVector3(maxx+minx, maxy+miny, maxz+minz);
    printf("Center Vector: %f, %f, %f\n", offset->X, offset->Y, offset->Z);
    *offset /= 2;
    for (int i = 0; i < Mesh->Vertices.size(); i++) {
       Mesh->Vertices.at(i).Position -= *offset;
    }

    
    //CMesh::resizeMesh(Mesh);
    float maxDistance = 0;
    for (int i = 0; i < Mesh->Vertices.size(); i++) {
       maxDistance = std::max(maxDistance, Mesh->Vertices.at(i).Position.length());
    }
    for (int i = 0; i < Mesh->Vertices.size(); i++) {
       Mesh->Vertices.at(i).Position /= maxDistance;
    } 

   //Fill arrays with vertices and indecies
   TriangleCount = Mesh->Triangles.size();

   float verts [Mesh->Vertices.size()*3];
   for (int i = 0; i < Mesh->Vertices.size(); i++) {
     verts[(i*3)]   = Mesh->Vertices.at(i).Position.X;
	  verts[(i*3)+1] = Mesh->Vertices.at(i).Position.Y;
     verts[(i*3)+2] = Mesh->Vertices.at(i).Position.Z;
   }

   unsigned int tris [Mesh->Triangles.size()*3];
   for (int i = 0; i < Mesh->Triangles.size(); i++) {
      tris[(i*3)]   = Mesh->Triangles.at(i).VertexIndex1;
      tris[(i*3)+1] = Mesh->Triangles.at(i).VertexIndex2;
      tris[(i*3)+2] = Mesh->Triangles.at(i).VertexIndex3;
      //printf("tri %d:  %d, %d, %d\n", i, tris[i*3], tris[i*3+1], tris[i*3+2]);
   }

   //compute normals per vertex
   float normals[Mesh->Vertices.size()*3];
   for (int i = 0; i < Mesh->Vertices.size()*3; i++) {
      normals[i] = 0;
   }
   
   float v1x, v2x, v3x, v1y, v2y, v3y, v1z, v2z, v3z;
   for (int i = 0; i < Mesh->Triangles.size(); i++) {
      v1x = Mesh->Vertices.at(Mesh->Triangles.at(i).VertexIndex1).Position.X;
      v1y = Mesh->Vertices.at(Mesh->Triangles.at(i).VertexIndex1).Position.Y;
      v1z = Mesh->Vertices.at(Mesh->Triangles.at(i).VertexIndex1).Position.Z;
      v2x = Mesh->Vertices.at(Mesh->Triangles.at(i).VertexIndex2).Position.X;
      v2y = Mesh->Vertices.at(Mesh->Triangles.at(i).VertexIndex2).Position.Y;
      v2z = Mesh->Vertices.at(Mesh->Triangles.at(i).VertexIndex2).Position.Z;
      v3x = Mesh->Vertices.at(Mesh->Triangles.at(i).VertexIndex3).Position.X;
      v3y = Mesh->Vertices.at(Mesh->Triangles.at(i).VertexIndex3).Position.Y;
      v3z = Mesh->Vertices.at(Mesh->Triangles.at(i).VertexIndex3).Position.Z;
      glm::vec3 v1 = glm::vec3(v1x-v2x, v1y-v2y, v1z-v2z);
      glm::vec3 v2 = glm::vec3(v1x-v3x, v1y-v3y, v1z-v3z);
      glm::vec3 cross = glm::cross(v1, v2);
      normals[Mesh->Triangles.at(i).VertexIndex1*3] += cross.x;
      normals[Mesh->Triangles.at(i).VertexIndex1*3+1] += cross.y;
      normals[Mesh->Triangles.at(i).VertexIndex1*3+2] += cross.z;      
      normals[Mesh->Triangles.at(i).VertexIndex2*3] += cross.x;
      normals[Mesh->Triangles.at(i).VertexIndex2*3+1] += cross.y;
      normals[Mesh->Triangles.at(i).VertexIndex2*3+2] += cross.z; 
      normals[Mesh->Triangles.at(i).VertexIndex3*3] += cross.x;
      normals[Mesh->Triangles.at(i).VertexIndex3*3+1] += cross.y;
      normals[Mesh->Triangles.at(i).VertexIndex3*3+2] += cross.z; 
   }
   float length;
   for (int i = 0; i < Mesh->Vertices.size(); i++) {
      length = sqrt( (normals[i*3]*normals[i*3]) + (normals[i*3+1]*normals[i*3+1]) + (normals[i*3+2]*normals[i*3+2]) );
      normals[i*3] /= length;
      normals[i*3+1] /= length;
      normals[i*3+2] /= length;
   }
   
   //pass vertices and indecies to graphics card
   glGenBuffers(1, &PositionBufferHandle);
   glBindBuffer(GL_ARRAY_BUFFER, PositionBufferHandle);
   glBufferData(GL_ARRAY_BUFFER, sizeof(verts), verts, GL_STATIC_DRAW);

   glGenBuffers(1, &TriBufferHandle);
   glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, TriBufferHandle);
   glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(tris), tris, GL_STATIC_DRAW);

   glGenBuffers(1, &VertexNormalBufferHandle);
   glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, VertexNormalBufferHandle);
   glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(normals), normals, GL_STATIC_DRAW);

   printf("success loading mesh: ");
   std::cout << fileName << std::endl;
	return true;
}

CMesh* const CMeshLoader::loadASCIIMesh(std::string const & fileName)
{
	std::ifstream File;
	File.open(fileName.c_str());

	if (! File.is_open())
	{
		std::cerr << "Unable to open mesh file: " << fileName << std::endl;
		return 0;
	}

	CMesh * Mesh = new CMesh();

	while (File)
	{
		std::string ReadString;
		std::getline(File, ReadString);

		std::stringstream Stream(ReadString);

		std::string Label;
		Stream >> Label;

		if (Label.find("#") != std::string::npos)
		{
			// Comment, skip
			continue;
		}

		if ("Vertex" == Label)
		{
			int Index;
			Stream >> Index; // We don't care, throw it away

			SVector3 Position;
			Stream >> Position.X;
			Stream >> Position.Y;
			Stream >> Position.Z;

			SVertex Vertex;
			Vertex.Position = Position;

			Mesh->Vertices.push_back(Vertex);
		}
		else if ("Face" == Label)
		{
			int Index;
			Stream >> Index; // We don't care, throw it away

			int Vertex1, Vertex2, Vertex3;
			Stream >> Vertex1;
			Stream >> Vertex2;
			Stream >> Vertex3;

			CMesh::STriangle Triangle;
			Triangle.VertexIndex1 = Vertex1 - 1;
			Triangle.VertexIndex2 = Vertex2 - 1;
			Triangle.VertexIndex3 = Vertex3 - 1;

			size_t Location;
			if ((Location = ReadString.find("{")) != std::string::npos) // there is a color
			{
				Location = ReadString.find("rgb=(");
				Location += 5; // rgb=( is 5 characters

				ReadString = ReadString.substr(Location);
				std::stringstream Stream(ReadString);
				float Color;
				Stream >> Color;
				Triangle.Color.Red = Color;
				Stream >> Color;
				Triangle.Color.Green = Color;
				Stream >> Color;
				Triangle.Color.Blue = Color;
			}

			Mesh->Triangles.push_back(Triangle);
		}
		else if ("" == Label)
		{
			// Just a new line, carry on...
		}
		else if ("Corner" == Label)
		{
			// We're not doing any normal calculations... (oops!)
		}
		else
		{
			std::cerr << "While parsing ASCII mesh: Expected 'Vertex' or 'Face' label, found '" << Label << "'." << std::endl;
		}
	}

	if (! Mesh->Triangles.size() || ! Mesh->Vertices.size())
	{
		delete Mesh;
		return 0;
	}

	return Mesh;
}



CMesh::CMesh()
{}

CMesh::~CMesh()
{}

void CMesh::centerMeshByAverage(CMesh &Mesh) {
}
void CMesh::centerMeshByExtents(SVector3 const &CenterLocation)
{
//ADD
}

void CMesh::resizeMesh(SVector3 const &Scale)
{
//ADD
}
