struct Material {
  vec3 aColor;
  vec3 dColor;
  vec3 sColor;
  float shine;
};

attribute vec3 aPosition;
attribute vec3 aNormal;

uniform mat4 uProjMatrix;
uniform mat4 uViewMatrix;
uniform mat4 uModelMatrix;

uniform vec3 uLightPos;
uniform vec3 uLightDir;
uniform vec3 uLColor;
uniform float uBrightness;
uniform float uGodView;

uniform vec3 uFlarePos;
uniform float uFlareBrightness;

uniform Material uMat;

varying vec4 vColor;
varying vec3 vNormal;
varying vec3 vWPosition;

void main() {
  vec4 Dl;
  
  vec4 Sl;
  vec3 normal; // holds vNormal as a normalized vec4
  vec3 temp;  // a temporary place for our vec3 to vec4 conversions  
  
  float brightness = (uGodView == 1.0) ? 10.0 : uBrightness;
  float innerRing = .14; // 8 degrees
  float outerRing = .35; // 20 degrees
  float maxRing = 1.2; 
  float fadeScale; // light fades when farther away from center of light
  
  float attenuation; // factor to multiply flashlight light by for distance fade
  float angleFromViewVector; // Angle between view vector and fragment position. use for spotlight
  
  //Remember to normalize the normal, as the linear interpolation of it is not unit length
  normal = normalize(vNormal)*-1.0;
  //Calculate attenuation based on distance between light origin and fragment 
  attenuation = 1.0/(pow(distance(vWPosition, uLightPos), 2.0));
  //Calculate angle in rads between (view vector) and (vector from light to fragment)
  angleFromViewVector = acos(dot(normalize(vWPosition-uLightPos), normalize(uLightDir)));
  
  // A simple calculation for our flare at the goal
  float distanceToFlare = distance(uFlarePos, vWPosition);
  vec4 flareLight;
  if (distanceToFlare < 5.0) {
     flareLight = vec4(0.0, 1.0/pow(distanceToFlare, 7.0), 0.0, 0.0) * uFlareBrightness;
  } else {
     flareLight = vec4(0.0, 0.0, 0.0, 0.0);
  }
  
  // if the angle between the vector from the light to the fragment position is in the outer ring
  // then create an outer light
  if (angleFromViewVector < outerRing) {
    //Diffuse Lighting = Color_of_light * max(dot(normal, light_source), 0.0) * color_of_object
    temp = uLColor * max(dot(normal, uLightDir), 0.0) * uMat.dColor;
    Dl = vec4(temp, 1.0);

    //Specular Lighting = Color_of_light * dot(view_vector, reflected_vector)^shininess * Color_of_model_shine
    vec3 reflected = reflect(uLightPos, vNormal);
    reflected = normalize(reflected);
    temp = uLColor * pow(max(dot(uLightDir, reflected), 0.0), uMat.shine) * uMat.sColor; 
    Sl = vec4(temp, 1.0);
    
    // if the angle is outside of the inner light ring
    // then make the light dimmer
    if (angleFromViewVector > innerRing) {
      // our fade scale is linear, from 1.0 to 0 from inner/outer gap to max outer ring
      fadeScale = (outerRing - angleFromViewVector) / (outerRing - innerRing);
      Dl = Dl * fadeScale*fadeScale;
      Sl = Sl * fadeScale*fadeScale;
    }
     
    gl_FragColor = ((Dl + vec4(uMat.aColor, 1.0)) * attenuation * 5.0 * brightness) + flareLight;
  // else we add ambient light in our forward direction  
  } else if (angleFromViewVector < maxRing) {
    fadeScale = (maxRing - angleFromViewVector) / (maxRing - innerRing);
    fadeScale = pow(fadeScale, 2.0);
    gl_FragColor = (vec4(uMat.aColor, 1.0) * attenuation * fadeScale * brightness) + flareLight;
  } else {
    if (uGodView == 0.0)
      discard;
    else
      gl_FragColor = vec4(uMat.aColor, 1.0) + flareLight;
  }
}
